/**
 * \file sg_path.hxx
 * Routines to abstract out path separator differences between MacOS
 * and the rest of the world.
 */

// Written by Curtis L. Olson, started April 1999.
//
// Copyright (C) 1999  Curtis L. Olson - http://www.flightgear.org/~curt
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// $Id: sg_path.hxx,v 1.11 2006-03-08 18:16:08 mfranz Exp $


#ifndef _SG_PATH_HXX
#define _SG_PATH_HXX

#include <sys/types.h>

#include <simgear/compiler.h>
#include STL_STRING

#include <simgear/math/sg_types.hxx>
#include <simgear/debug/logstream.hxx>

SG_USING_STD(string);

#ifdef _MSC_VER
  typedef int mode_t;
#endif

/**
 * define directory path and search separators
 */

#if defined( macintosh )
#define SGPATH_sgDirPathSep ':'
#define SGPATH_sgDirPathSepBad '/'
#else
#define SGPATH_sgDirPathSep '/'
#define SGPATH_sgDirPathSepBad '\\'
#endif

#if defined( WIN32 ) && !defined(__CYGWIN__)
#define SGPATH_sgSearchPathSep ';'
#else
#define SGPATH_sgSearchPathSep ':'
#endif


/**
 * Split a directory string into a list of it's parent directories.
 */
static inline string_list sgPathBranchSplit( const string &dirpath ) {
    string_list path_elements;
    string element, path = dirpath;
    while ( path.size() ) {
        size_t p = path.find( SGPATH_sgDirPathSep );
        if ( p != string::npos ) {
            element = path.substr( 0, p );
            path.erase( 0, p + 1 );
        } else {
            element = path;
            path = "";
        }
        if ( element.size() )
            path_elements.push_back( element );
    }
    return path_elements;
}


/**
 * Split a directory search path into a vector of individual paths
 */
static inline string_list sgPathSplit( const string &search_path ) {
    string tmp = search_path;
    string_list result;
    result.clear();

    bool done = false;

    while ( !done ) {
        int index = tmp.find(SGPATH_sgSearchPathSep);
        if (index >= 0) {
            result.push_back( tmp.substr(0, index) );
            tmp = tmp.substr( index + 1 );
        } else {
            if ( !tmp.empty() )
                result.push_back( tmp );
            done = true;
        }
    }

    return result;
}


/**
 * A class to hide path separator difference across platforms and assist
 * in managing file system path names.
 *
 * Paths can be input in any platform format and will be converted
 * automatically to the proper format.
 */

class SGPath {

private:

    string path;

public:

    /** Default constructor */
    SGPath();

    /**
     * Construct a path based on the starting path provided.
     * @param p initial path
     */
    SGPath( const string& p ) : path(p)
    {
        fix();
    }

    /** Destructor */
    ~SGPath() {}

    /**
     * Set path to a new value
     * @param p new path
     */
    void set( const string& p );
    SGPath& operator= ( const char* p ) { this->set(p); return *this; }

    /**
     * Append another piece to the existing path.  Inserts a path
     * separator between the existing component and the new component.
     * @param p additional path component */
    void append( const string& p ) {
        if ( path.size() == 0 ) {
            path = p;
        } else {
            if ( p[0] != SGPATH_sgDirPathSep ) {
                path += SGPATH_sgDirPathSep;
            }
            path += p;
        }
        fix();
    }

    /**
     * Append a new piece to the existing path.  Inserts a search path
     * separator to the existing path and the new patch component.
     * @param p additional path component */
    void add( const string& p );

    /**
     * Concatenate a string to the end of the path without inserting a
     * path separator.
     * @param p addtional path suffix
     */
    void concat( const string& p ) {
        if ( path.size() == 0 ) {
            path = p;
        } else {
            path += p;
        }
        fix();
    }

    /**
     * Get the file part of the path (everything after the last path sep)
     * @return file string
     */
    string file() const;
  
    /**
     * Get the directory part of the path.
     * @return directory string
     */
    string dir() const {
        int index = path.rfind(SGPATH_sgDirPathSep);
        if (index >= 0) {
            return path.substr(0, index);
        } else {
            return "";
        }
    }
  
    /**
     * Get the base part of the path (everything but the extension.)
     * @return the base string
     */
    string base() const;

    /**
     * Get the extention part of the path (everything after the final ".")
     * @return the extention string
     */
    string extension() const;

    /**
     * Get the path string
     * @return path string
     */
    string str() const { return path; }

    /**
     * Get the path string
     * @return path in "C" string (ptr to char array) form.
     */
    const char* c_str() { return path.c_str(); }

    /**
     * Determine if file exists by attempting to fopen it.
     * @return true if file exists, otherwise returns false.
     */
    bool exists() const;

    /**
     * Create the designated directory.
     */
    void create_dir(mode_t mode)
    {
        string_list dirlist = sgPathSplit(dir());
        if ( dirlist.empty() )
            return;
        string path = dirlist[0];
        string_list path_elements = sgPathBranchSplit(path);
        bool absolute = !path.empty() && path[0] == SGPATH_sgDirPathSep;

        unsigned int i = 1;
        SGPath dir = absolute ? string( 1, SGPATH_sgDirPathSep ) : "";
        dir.concat( path_elements[0] );
    #if defined( _MSC_VER) || defined(__MINGW32__)
        if ( dir.str().find(':') != string::npos && path_elements.size() >= 2 ) {
            dir.append( path_elements[1] );
            i = 2;
        }
    #endif
        struct stat info;
        int r;
        for(; ( r = stat( dir.c_str(), &info ) ) == 0 && i < path_elements.size(); i++) {
            dir.append(path_elements[i]);
        }
        if ( r == 0 ) {
            return; // Directory already exists
        }
        #if defined( _MSC_VER) || defined(__MINGW32__)
        #  define sgMkDir(d,m)       _mkdir(d)
        #else
        #  define sgMkDir(d,m)       mkdir(d,m)
        #endif
        if ( sgMkDir( dir.c_str(), mode) ) {
            SG_LOG( SG_IO, SG_ALERT, "Error creating directory: " + dir.str() );
            return;
        }
        for(;i < path_elements.size(); i++) {
            dir.append(path_elements[i]);
            if ( sgMkDir( dir.c_str(), mode) ) {
                SG_LOG( SG_IO, SG_ALERT, "Error creating directory: " + dir.str() );
                break;
            }
        }
        #undef sgMkDir
    }

private:

    // If Unix, replace all ":" with "/".  If MacOS, replace all "/" with
    // ":" it should go without saying that neither of these characters
    // should be used in file or directory names.  In windoze, allow the
    // second character to be a ":" for things like c:\foo\bar

    void fix()
    {
        for ( string::size_type i = 0; i < path.size(); ++i ) {
    #if defined( WIN32 )
            // for windoze, don't replace the ":" for the second character
            if ( i == 1 ) {
                continue;
            }
    #endif
            if ( path[i] == SGPATH_sgDirPathSepBad ) {
                path[i] = SGPATH_sgDirPathSep;
            }
        }
    }

};

#endif // _SG_PATH_HXX


